/* A SVGAlib Web browser "wb0"
   image_parser.c: parses incoming image data
   Supports PCX, truecolor PCX, GIF, interlaced GIF, JPEG, PNG,
   interlaced PNG, alpha PNG
   (c) 1999 Clocksoft -- Karel Kulhavy. clock@atrey.karlin.mff.cuni.cz
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <jpeglib.h>
#include <png.h>

#include "gvars.h"
#include "scroller.h"
#include "image_parser.h"
#include "line_breaker.h"
#include "block.h"

//#define DG_IMAGE_PARSER
//#define DG_CODES
//#define DG_PIXELS
//#define DG_FLUSH
//#define DG_ERR
#define DG_GAMMA

#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))
#define RND(x) (floor(.5+(x)))
#define MULDIVRND(c,n,d) (((c)*(n)+(n/2))/(d))
#define BTOP(c) MULDIVRND(PRESCALE,c,255)
#define WTOP(c) MULDIVRND(PRESCALE,c,65535)

/******************* Internal variables ***************************************/

/* State of automatus finitus recognizing the image formats */
static int state=0;

/* Image width, height, bits per pixel, byte per line, number of bit planes */
static int im_width;
static int im_height;
static int im_bpp;
static int im_bpl;
static int im_planes;

static unsigned char *raw_line;
static unsigned char *raw_line_end;
static unsigned char *raw_line_ptr;
static unsigned char *big_buffer;

/* Actual line with truecolor values to be output*/
static int *actual_line;

/* For Floyd-steinberg */
static int *error_line; /* Is not allocated for truecolor vga mode
                           0 means no error, PRESCALE means plus white error,
                           -PRESCALE means minus white error */
static int line_width; /* Length of this and actual_line in pixels. */

/* Temporary buffer for remembreing headers and so on. */
static unsigned char tbuf[128];

/* tbuf length */
static int tlen=0;

/* Offsets in image, shift with every pixel received */
static int xoff, yoff;

/* Used for keying out the unwanted block counts in raster data. */
static int remains;

/* Actual code size for decoding */
static int code_size;

/* Initial code size delivered by coder. */
static int initial_code_size;

/* Number of bits already loaded of a codesize-long code. */
static int bits_read;

/* Read code in accept_byte */
static unsigned long read_code;

/* CC */
static int CC;

/* EOI code */
static int EOI;

/* Last code for accept_code() routine */
static int last_code;

/* First free position in the decoder table */
static int table_pos;

/* Flag: first code is treated a little bit differetly than the following */
static int first_code;

/* There is an internal representation for each index for colormapped
 * pictures. Length=3*number of entries. PRESCALE
 */
static int *color_map;

/* First usable line, line at which starts the image */
static int usable_line, image_line;

/* error-occured flag for setjmp/longjmp. */
static int error_occured;

struct table_entry
{
 unsigned char end_char;
 unsigned char garbage; /* This has nothing common to do with code table:
                           this is temporarily used for reverting strings :-) */

 short pointer; /* points onto another entry in table, number 0...4095.
                   number -1 means it end there, the end_char is the last
		   number -2 means that this entry is no filled in yet.
                */
		   
};

/* Table for storing dictionary during decoding */
static struct table_entry table[4096];

/* For interlacd GIFs, the distance between consecutive lines.
   But because of two first passes have both distance=8, the first will
   have distance=24 and the 16 will be masked out by 0xf.
*/
static int interl_dist;

/* Flag, whether in PCX is a 256-color palette at the end of the file. */
static int palette_present;

/* The number how many times repeat the byte in run length encoding */
static int run_length;

/* A cyclic buffer for accepting bytes.
   At the end of the file, it is checked, whether there were at least 769 bytes,
   the 12 at the beginning is tested and palette set.
*/
static unsigned char *cyclic=NULL;

/* The first free byte in cyclic. */
static int cyclic_pos;

/* Flag telling that there were at least 769 bytes going into cyclic buffer. */
static int cyclic_enough;

/* For JPEG decoding */
static struct jpeg_decompress_struct cinfo;
static struct 
{
 struct jpeg_error_mgr pub;
 jmp_buf setjmp_buffer;
}jerr;

/* JPEG input data buffer */
unsigned char *jdata=NULL;

/* Only for JPEG decoding.
  Indicates, how many data should be skipped.
*/
long skip_bytes;

/* PNG structures */
png_structp png_ptr;
png_infop info_ptr;

/* My PNG flag */
int png_complete;

/****************************** Functions *************************************/

int
issue_image_msg(const unsigned char *string)
{
 struct lineno_desc b;

 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"issue_image_msg(%s)\n",string);
 #endif
 b.ptr=(unsigned char *)string;
 b.number=strlen(string);
 b.lineno=usable_line+im_height;
 usable_line+=font_height;
 if (scroller(SC_ADD_LINENO,&b))
 {
  fprintf(stderr,"Can't issue warning\n");
  return 1;
 }
 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"!issue_image_msg\n");
 #endif
 return 0;
}

int
alloc_color_map(int colors)
{
 free(color_map);
 color_map=malloc(colors*3*sizeof(*color_map));
 return !color_map;
}

/* r,g,b : PRESCALE */
void
setup_color_map(int index, int r, int g, int b)
{
 color_map[3*index]=r;
 color_map[3*index+1]=g;
 color_map[3*index+2]=b;
}

/* Make a default color map for PCX and for GIFs where no colormap is included.
 * Generates a palette which is probably not the right one.
 */
void
default_color_map(void)
{
 int c,r,g,b;
 
 alloc_color_map(256);
 for (c=0;c<256;c++)
 {
  double angle=c*M_PI/128;
  r=RND(PRESCALE/2+PRESCALE/2*sin(angle));
  g=RND(PRESCALE/2+PRESCALE/2*sin(angle*2));
  b=RND(PRESCALE/2+PRESCALE/2*cos(angle*3));
  setup_color_map(c,r,g,b);
 }
}

/* width*3 integers for Floyd-Steinberg dithering errors (error_line)
 * width*3 integers for actual line (actual_line)
 */
void
alloc_line_buffers(int width)
{
 line_width=width;
 free(actual_line);
 actual_line=xalloc(line_width*3*sizeof(*actual_line)); /* Truecolor always */
 free(error_line);
 error_line=xcalloc(line_width*3*sizeof(*error_line)); /* May be negative or positive */
}

void
alloc_big_buffer(int length)
{
 free(big_buffer);
 big_buffer=xalloc(length);
}

void
free_big_buffer(void)
{
 free(big_buffer);
 big_buffer=NULL;
}

void
free_line_buffers(void)
{
 free(actual_line);
 actual_line=NULL;
 free(error_line);
 error_line=NULL;
}

void inline
add_error(unsigned char *dest, int src)
{
 int a;
 a=*dest+src;
 if (a<0) a=0;
 if (a>255) a=255;
 *dest=a;
}

void
zero_error_line(void)
{
 if (error_line)
  bzero(error_line,line_width*3*sizeof(*error_line));
}

/* Organization of the actulal_line: rgbrgbrgb...
   Organization of errror_line: rgbrgbrgbrgb...
   Dithers even truecolor pictures.
   Destroys the actual_buffer.
   Takes yoff as line number.
 */
int
dither_and_flush(void)
{
 int x;
 unsigned char *output=xalloc(line_width*vga_bytes);
 int r,g,b,r0,g0,b0;
 
 for (x=0;x<line_width*3;x++)
 {
  actual_line[x]+=error_line[x];
 }
 zero_error_line();
 x=0;
 {
  r=actual_line[3*x];
  g=actual_line[3*x+1];
  b=actual_line[3*x+2];
  round_pixel(output+x*vga_bytes,&r,&g,&b);
  r0=(r+8)>>4;
  g0=(g+8)>>4;
  b0=(b+8)>>4;
  r-=9*r0;
  g-=9*g0;
  b-=9*b0;
  error_line[3*x+3]+=r0;
  error_line[3*x+4]+=g0;
  error_line[3*x+5]+=b0;
  error_line[3*x]+=r0*5;
  error_line[3*x+1]+=g0*5;
  error_line[3*x+2]+=b0*5;
  actual_line[3*x+3]+=r;
  actual_line[3*x+4]+=g;
  actual_line[3*x+5]+=b;
 }
 for (x=1;x<line_width-1;x++)
 {
  r=actual_line[3*x];
  g=actual_line[3*x+1];
  b=actual_line[3*x+2];
  round_pixel(output+x*vga_bytes,&r,&g,&b);
  r0=(r+8)>>4;
  g0=(g+8)>>4;
  b0=(b+8)>>4;
  r-=9*r0;
  g-=9*g0;
  b-=9*b0;
  error_line[3*x+3]+=r0;
  error_line[3*x+4]+=g0;
  error_line[3*x+5]+=b0;
  error_line[3*x]+=5*r0;
  error_line[3*x+1]+=5*g0;
  error_line[3*x+2]+=5*b0;
  error_line[3*x-3]+=3*r0;
  error_line[3*x-2]+=3*g0;
  error_line[3*x-1]+=3*b0;
  actual_line[3*x+3]+=r;
  actual_line[3*x+4]+=g;
  actual_line[3*x+5]+=b;
 }
 x=line_width-1;
 {
  r=actual_line[3*x];
  g=actual_line[3*x+1];
  b=actual_line[3*x+2];
  round_pixel(output+x*vga_bytes,&r,&g,&b);
  r0=(r+8)>>4;
  g0=(g+8)>>4;
  b0=(b+8)>>4;
  error_line[3*x]+=5*r0;
  error_line[3*x+1]+=5*g0;
  error_line[3*x+2]+=5*b0;
  error_line[3*x-3]+=3*r0;
  error_line[3*x-2]+=3*g0;
  error_line[3*x-1]+=3*b0;
 }
 {
  struct lineno_desc b={yoff+image_line,line_width,output};
  if (scroller(SC_ADD_MICRO_LINENO,&b)) goto err;
 }
 free(output);
 return 0;
 err:
 free(output);
 fprintf(stderr,"Can't dither and output line.\n");
 return 1;
}

void
free_color_map(void)
{
 free(color_map);
 color_map=NULL;
}

/* GIF code.
   Initialize code table: construct codes 0...(1<<code_size)-1 with values
   0...(1<<code_size)-1 Codes (1<<code_size) and (1<<code_size)+1 are
   left intact -- they are of no use.
   Initializes CC and EOI
   Zeroes out the last_code. In normal data stream the first code must be
   in the table. However, if corrupted stream is to be received, a cause could
   happen that the first code would be out of table and as last code would
   be used something uninitialized and something very strange could happen
   (drawing a pixel from previous image or an infinite loop in outputting
   string)
*/
void
init_table(void)
{
 int i;
 first_code=1;
 code_size=initial_code_size;
 for (i=0;i<1<<code_size;i++)
 {
  table[i].pointer=-1;
  table[i].end_char=i;
 }
 /* Here i=1<<code_size. */
 CC=i;
 EOI=i+1;
 table_pos=i+2;
 for (;i<4096;i++)
 {
  table[i].pointer=-2; 
 }
 code_size++;
 last_code=0;
}

/* Spreads yoff-th line in big_buffer into n-1 following lines. */
void
gif_line_spread_in_big_buffer(int n)
{
 int y;
 
 if (yoff+n>im_height) n=im_height-yoff;
 for (y=yoff+1;y<yoff+n;y++)
  memcpy(big_buffer+im_width*y,big_buffer+im_width*yoff,im_width);
}

/* Using yoff fetches the line from big_buffer into actual_line. */
void
fetch_gif_actual_line(void)
{
 int c,x;
 
 for (x=0;x<im_width;x++)
 {
  c=3*big_buffer[yoff*im_width+x];
  actual_line[x*3]=color_map[c];
  actual_line[x*3+1]=color_map[c+1];
  actual_line[x*3+2]=color_map[c+2];
 }
}

void
apply_gamma(void)
{
 int x;

 if (image_gamma==1||image_gamma<=0) return;

 for (x=0;x<line_width*3;x++)
 {
  int v=actual_line[x];
  if (v<0) v=0;
  actual_line[x]=RND(PRESCALE*pow((double)v/PRESCALE,1/image_gamma));
 }
}

/* Increaes yoff by some value. Used in interlaced GIFs only. */
void
dither_and_flush_line_block_from_big_buffer(int n)
{
 int end=yoff+n;

 if (end>im_height) end=im_height;
 
 for (;yoff<end;yoff++)
 {
  fetch_gif_actual_line();
  /* For truecolor, no dithering is performed. */
  apply_gamma();
  dither_and_flush();
 }
}

/* GIF code
   Outputs a single pixel.
   Return value: 0 ok
                 1 fatal error
		 2 stop sending image data
*/
int
output_pixel(int c)
{
 #ifdef DG_PIXELS
  putchar(c);
 #endif
 if (interl_dist!=1)
 {
  big_buffer[im_width*yoff+xoff]=c;
 }
 actual_line[xoff*3]=color_map[c*3];
 actual_line[xoff*3+1]=color_map[c*3+1];
 actual_line[xoff*3+2]=color_map[c*3+2];
 xoff++;
 if (xoff>=im_width)
 {
  xoff=0;
  if (interl_dist==1)
  { 
   apply_gamma();
   dither_and_flush();
   yoff++;
  }
  else
  {
   gif_line_spread_in_big_buffer((interl_dist==24)?8:(interl_dist>>1));
   dither_and_flush_line_block_from_big_buffer(interl_dist&0xf);
  }
  if (yoff>=im_height)
  {
   /* The vertical range is complete. */
   if (interl_dist<=2)
   {
    /* The image is ready, don't send any further data. */
    return 2;
   }
   else
   {
    if (interl_dist==24)
    {
     interl_dist=8;
    }
    else
    {
     interl_dist>>=1;
    }
    zero_error_line();
    yoff=0;
    dither_and_flush_line_block_from_big_buffer(interl_dist>>1);
    yoff=interl_dist>>1;
   }
  }
 }
 return 0;
}

/* Finds the first char of output string for given codeword. */
int
find_first(int c)
{
 int p;
 int i;
 for (i=0;i<4096;i++)
 {p=table[c].pointer;
  if (p==-1) break;
  if (p<-1||p>=4096) return 0;
  c=p;
 }
 return table[c].end_char;
}

/* GIF code
   Supply a code and it outputs the string for c. 
   Return value: 0 ok
                 1 fatal error
		 2 stop sending data
*/
int
output_string(int c)
{
 int pos=0;
 
 while(1)
 {
  if (pos==4096)
  {
   /* Evident error. This is fuse for 4096 buffer overflow and against cycles */
   if (issue_image_msg("Evident error in GIF codestream: infinitely cyclic"
    " string.\n"))
   {
    fprintf(stderr,"Unable to issue error.\n");
    return 1;
   }
   return 2;
  }
  table[pos].garbage=table[c].end_char;
  pos++;
  c=table[c].pointer;
  if (c<0) break; /* We are at the end */
 }
 for (pos--;pos>=0;pos--)
 {switch (output_pixel(table[pos].garbage))
  {
   case 1:
   fprintf(stderr,"Unable to output pixel.\n");
   return 1;

   case 2:
   return 2; /* Stop sending data */
  }
 }
 return 0;
}
/* Adds to the code table
   return value: 0 ok
                 1 fatal error
		 2 stop sending data
*/

int
add_table(int end_char,int pointer)
{
 /* First check overflow */
 if (table_pos>=4096)
 {
  /* Overflow */
  if (issue_image_msg("Bad GIF: table overflow happened."))
  {
   fprintf(stderr,"Can't issue table overflow error.\n");
   return 1;
  }
  return 2; /* Stop sending data */
 }
 
 table[table_pos].end_char=end_char;
 table[table_pos].pointer=pointer;
 table_pos++;
 if (table_pos==(1<<code_size)&&code_size!=12)
 {
  /* Table pos is a power of 2 */
  code_size++;
  #ifdef DG_CODES
   printf("inc\n");
  #endif
 }
 return 0;
}

/* Yout throw inside a codeword and it processes it farther
   Return value: 0 ok
                 1 fatal error
		 2 stop sending data
   If the code==256, it means that end-of-file came.
   This is part of GIF code.
*/
int
accept_code(int c)
{
 int k=0; /* 0 to suppress warning */

 #ifdef DG_CODES
  printf("%i\n",c);
 #endif
 if (c>4096||c<0)
 {
  /* An erroneous code word came, ignore. */
  return 0;
 }
 if (c==CC)
 {
  #ifdef DG_IMAGE_PARSER
   printf("<CC> received.\n");
  #endif
  init_table();
  return 0;
 }
 
 if (c==EOI)
 {
  #ifdef DG_IMAGE_PARSER
   printf("<EOI> received.\n");
  #endif
  return 2; /* Don't send further data */
 }
 
 if (first_code)
 {
  first_code=0;
  /* First code after init_table */
  /* Action: output the tring for code */
  switch(output_string(c))
  {
   case 1:
   fprintf(stderr,"Unable to output string for forst code after table_init.\n");
   return 1;
  
   case 2:
   return 2; /* Stop sending data */
  }
  last_code=c;
  return 0;
 }
 
 if (c>=table_pos)
 {
  /* The code is not in the table */
  k=find_first(last_code);
 }
 else
 {
  /* The code is in code table */
  k=find_first(c);
 }
  
 switch (add_table(k,last_code))
 {case 1:
  fprintf(stderr,"Unable to add to the table.\n");
  return 1;

  case 2:
  return 2; /* No more data wanted */
 }
 
 /* Output the string for code */
 switch (output_string(c))
 {
  case 1: /* Fatal error */
  fprintf(stderr,"Unable to output string.\n");
  return 1;

  case 2: /* Stop sending data */
  return 2;
 }
 last_code=c; /* Update last code. */
 return 0;
}


/* You throw inside a byte, it depacks the bits out and makes code and then
   passes to the decoder (no headers, palettes etc. go through this.)
   Return: 0 ok.
           1 fatal error
	   2 stop sending data
*/
int
accept_byte(unsigned char c)
{
 int original_code_size;
 
 read_code|=(unsigned long)c<<bits_read;  
 bits_read+=8;
 while (bits_read>=code_size)
 {
  /* We have collected up a whole code word. */
  original_code_size=code_size;
  switch(accept_code(read_code&((1<<code_size)-1)))
  {
   case 1:
   fprintf(stderr,"Unable to accept code word.\n");
   return 1;

   case 2:
   return 2;
  }
  read_code>>=original_code_size;
  bits_read-=original_code_size;
 }
 return 0;
}

/* Takes yoff and places raw_line on yoff-the row of big_buffer */
void
add_pcx_raw_line(void)
{
 memcpy(big_buffer+im_bpl*im_planes*yoff,raw_line,im_bpl*im_planes);
}

/* Takes yoff and places raw_line on yoff-the row of big_buffer */
void
fetch_pcx_raw_line(void)
{
 memcpy(raw_line,big_buffer+im_bpl*im_planes*yoff,im_bpl*im_planes);
}

/* Takes the raw_line and im_bpl and im_planes and im_bpp and deducts how to
   display the contents of the raw_line and displays it.
   It always displays only one micro line.
*/
int
flush_pcx_line(void)
{
 int width,a;
 
 width=min(im_bpl,im_width);
 if (im_bpp==8&&im_planes==3)
 { 
  /* Looks like truecolor */
  for (a=0;a<width;a++)
  {
   actual_line[a*3]=BTOP(raw_line[a]);
   actual_line[a*3+1]=BTOP(raw_line[a+im_bpl]);
   actual_line[a*3+2]=BTOP(raw_line[a+2*im_bpl]);
  }
  if (apply_gamma(),dither_and_flush()) goto err;
 }
 else if (im_bpp==8&&im_planes==1)
 {
  int a;

  /* Looks like 256-color (colormapped) */
  for (a=0;a<width;a++)
  {
   int c=raw_line[a];
   actual_line[a*3]=color_map[3*c];
   actual_line[a*3+1]=color_map[3*c+1];
   actual_line[a*3+2]=color_map[3*c+2];
  }
  if (apply_gamma(),dither_and_flush()) goto err;
 }
 else
 {
  /* Now let's assume that im_bpp==1 */
  int plane,bit,shift=7;
  unsigned char *inptr;
  unsigned long value;
  int number;
  
  number=min(im_width,im_bpl*8);
  inptr=raw_line;
  for (bit=0;bit<number;bit++)
  {
   value=0;
   for (plane=0;plane<im_planes;plane++)
   {
    value|=((inptr[plane*im_bpl]>>shift)&1)<<plane;
   } 
   actual_line[bit*3]=color_map[3*value];
   actual_line[bit*3+1]=color_map[3*value+1];
   actual_line[bit*3+2]=color_map[3*value+2];
   if (!shift)
   {
    shift=7;
    inptr++;
   }
   else
   {
    shift--;
   }
  }
  if (apply_gamma(),dither_and_flush()) goto err;
 }
 return 0;
 err:
 fprintf(stderr,"Can't flush PCX line.\n");
 return 1;
}

/* length is in bytes. */
void
alloc_raw_line(int length)
{
 free(raw_line);
 raw_line=xalloc(length);
 raw_line_end=raw_line+length;
 raw_line_ptr=raw_line;
}

void
free_raw_line(void)
{
 free(raw_line);
 raw_line=NULL;
}

/* Call this every time you complete raw_line.
   return value: 0 ok
                 1 fatal error
		 2 stop sending data (but the palette is still possible
		                      to be received. )
                 3 really stop sending all data
*/
int
accept_pcx_eol()
{
 raw_line_ptr=raw_line;
 #ifdef DG_IMAGE_PARSER
  printf("raw_line full.\n");
 #endif
 /* Flush the raw_line */
 if (palette_present)
 {
  add_pcx_raw_line();
 }
 if (flush_pcx_line())
 {
  fprintf(stderr,"image parser: unable to flush PCX raw line.\n");
  return 1;
 }
 yoff++;
 #ifdef DG_IMAGE_PARSER
  printf("raw_line flushed.\n");
 #endif
 if (yoff>=im_height) /* Whole image read */
 {
  if (palette_present)
  {
   cyclic=(unsigned char *)malloc(769);
   if (!cyclic)
   {
    fprintf(stderr,"Unable to malloc 769 bytes for cyclic buffer for reading"
     " PCX 256-colour palette at the end of the file.\n");
    return 1;
   }
   cyclic_pos=0;
   cyclic_enough=0;
   return 2;
  }
  else
  {
   #ifdef DG_IMAGE_PARSER
    printf("Whole image read (I'm in accept_pcx_byte).\n");
   #endif
   free_line_buffers();
   free_raw_line();
   return 3; /* Everything is complete */
  }
 }
 return 0;
}

/* Takes cyclic buffer as is with cyclic_pos and performs all the necessary
   integrity tests and sets up the 256-color PCX palette.
   There should be 769 bytes in the cyclic buffer, cyclic_enough set
   and the first (the oldest) byte should be set to 12.
   return value: 0 ok
                 1 fatal error
*/
int
accept_raw_pcx_palette(void)
{
 int a;
 
 if (!cyclic_enough||cyclic[cyclic_pos]!=12)
 {
  /* Bad palette */
  if (issue_image_msg("Bad PCX palette."))
  {
   fprintf(stderr,"image_parser: unable to issue error about bad PCX"
    " palette.\n");
   return 1;
  }
  return 0;
 }
 /* The cyclic_pos is 0..768 here */
 alloc_color_map(256);
 for (a=cyclic_pos+1;a<769;a++)
 {
  color_map[a-cyclic_pos-1]=BTOP(cyclic[a]);
 }
 for (a=0;a<cyclic_pos;a++)
 {
  color_map[769-cyclic_pos-1+a]=BTOP(cyclic[a]);
 }
 return 0;
}

/* Takes 48 bytes of paldata and adds 256-16 random palette values.
   If the im_bpp*im_planes>4, then doesn't do the things with 16 colors
   and only sets up default (ugly) palette.
   return value: 0 ok
                 1 fatal error
*/
int
setup_16_pcx_palette(unsigned char *paldata)
{
 if (im_bpp*im_planes<=4)
 {
  int a;
  
  alloc_color_map(16);
  for (a=0;a<48;a++)
  {
   color_map[a]=BTOP(paldata[a]);
  }
 }
 else if (im_bpp*im_planes<=8)
 {
  default_color_map();
 }
 return 0;
}

/* Uses big buffer to redraw the PCX with the palette. */
int
redraw_pcx(void)
{
 zero_error_line();
 for (yoff=0;yoff<im_height;yoff++)
 {
  fetch_pcx_raw_line();
  if (flush_pcx_line()) goto err;
 }
 return 0;
 err:
 fprintf(stderr,"Can't redraw PCX.\n");
 return 1;
}

void
take_image_line(void)
{
 image_line=usable_line;
 usable_line+=im_height;
}

/* Takes up one byte from file.
   Return value:
   0=ok
   1=error
   2=stop sending data
   
   Input: 256 means EOF
*/
int inline
accept(int c)
{
 switch(state)
 {
  case 0: /* Reading signature and screen descriptor -- 13 bytes in GIF*/
          /* This is beginning of GIF code */
  tbuf[tlen]=c;
  tlen++;
  if (tlen>=13)
  {
   if (strncmp(tbuf,"GIF87a",6)&&strncmp(tbuf,"GIF89a",6))
   {
    /* Header is invalid */
    if (issue_image_msg("Invalid GIF signature."))
    {
     fprintf(stderr,"Can't issue error message.\n");
     return 1;
    }
    #ifdef DG_IMAGE_PARSER
     printf("Stop sending data! (parser)\n");
    #endif
    return 2; /* No more data wanted */
   }
   im_bpp=(tbuf[10]&7)+1;
   tlen=0;
   if (tbuf[10]<128)
   {
    /* No global color map follows, we will have to think out a default global
       color map */
    default_color_map(); /* Alloc color map with 256 entries */
    state=2; /* Reading image data */
   }
   else
   {
    alloc_color_map(1<<im_bpp);
    state=1; /* Reading palette in GIF */
   }
   #ifdef DG_IMAGE_PARSER
    printf("Signature read. bpp=%i, global color map=%i\n",im_bpp,2-state);
   #endif
  }
  break;

  case 1: /* Reading palette -- 3*(1<<im_bpp) bytes in GIF*/
  color_map[tlen]=BTOP(c);
  tlen++;
  if (tlen>=3*(1<<im_bpp))
  {
   #ifdef DG_IMAGE_PARSER
    printf("Palette of %i entries read.\n",1<<im_bpp);
   #endif
   state=2;
   tlen=0;
  }
  break;
  
  case 2: /* Reading garbage before and one ',' or '!' in GIF */
  switch (c)
  {
   case ',':
   state=3;
   #ifdef DG_IMAGE_PARSER
    printf("image parser: ',' separator read.\n");
   #endif
   tlen=0;
   goto end_of_2; 
   
   case '!':
   state=7;
   goto end_of_2;
  }
  #ifdef DG_IMAGE_PARSER
   printf("%c found, neither ',' nor '!', ignored.\n",c);
  #endif
  end_of_2:
  break;

  case 3: /* Reading image descriptor -- 9 bytes in GIF */
  tbuf[tlen]=c;
  tlen++;
  if (tlen>=9)
  {
   im_width=tbuf[4]+(tbuf[5]<<8);
   im_height=tbuf[6]+(tbuf[7]<<8);
   if (im_width<=0||im_height<=0)
   {
    if (issue_image_msg("Bad image dimensoins of GIF."))
    {
     fprintf(stderr,"Unable to issue err msg.\n");
     return 1;
    }
    return 2; /* Stop sending data */
   }
   take_image_line();
   alloc_line_buffers(im_width);
   #ifdef DG_IMAGE_PARSER
    printf("Image size %ix%i.\n",im_width,im_height);
   #endif
   if (tbuf[8]&128)
   {
    state=4;
    tlen=1<<(1+(tbuf[8]&7));
   }
   else
   {
    state=5;
    tlen=0;
    xoff=0;
    yoff=0;
   }
   if (tbuf[8]&64)
   {
    /* Interlaced order */
    interl_dist=24; /* Actually 8, the 16 indicates it is the first pass. */
    alloc_big_buffer(im_height*im_width);
   }
   else
   {
    interl_dist=1;
   }
  } 
  break;

  case 4: /* Skipping local colormap in GIF */
  tlen--;
  if (!tlen)
  {
   #ifdef DG_IMAGE_PARSER
    printf("Local color map skipped and ignored.\n");
   #endif
   state=5;
   xoff=0;
   yoff=0;
  }
  break;

  case 5: /* Reading code size block in GIF */
  initial_code_size=c;
  if (initial_code_size<=1||initial_code_size>8)
  {
   if (issue_image_msg("Invalid initial code size."))
   {
    fprintf(stderr,"Unable to issue error.\n");
    return 1;
   }
   free_line_buffers();
   free_big_buffer();
   return 2;
  }
  #ifdef DG_IMAGE_PARSER
   printf("Initial code size=%i.\n",code_size);
  #endif
  bits_read=0;
  read_code=0;  /* Nothing read */
  init_table(); /* Decoding is about to begin sets up code_size. */
  state=6;
  tlen=0;
  remains=0;
  break;

  case 6: /* Reading image data in GIF */
  if (!remains)
  {
   /* This byte is a count byte. Feed it into remains. */
   remains=c;
   if (!c)
   {
    #ifdef DG_IMAGE_PARSER
     printf("End of blocks mark encountered.\n");
    #endif
    /* 0 count = end of data. */
    if (xoff||yoff)
    {
     /* Notify user about the fact that the image was incomplete */
     /* If the image was complete, the ?pos would be both zero (they normally
        wrap at the end of line and at the end of the image */
     {
      /* Black out the remainder of the actual_line */
      int a;
      for (a=xoff*3;a<3*line_width;a++)
       actual_line[a]=0;
      #ifdef DG_IMAGE_PARSER
       printf("Blacking out remainder of last line.\n");
      #endif
     }
     if (apply_gamma(),dither_and_flush())
     {
      fprintf(stderr,"Unable to flush last line of incomplete image.\n");
      return 1;
     }
     #ifdef DG_IMAGE_PARSER
      printf("Issuing error about image incompleteness.\n");
     #endif
     if (issue_image_msg("Image is incomplete."))
     {
      fprintf(stderr,"Unable to issue error.\n");
      return 1;
     }
    }
    free_line_buffers();
    free_big_buffer();
    return 2; /* Don't send any following data */
   }
   else
   {
    #ifdef DG_IMAGE_PARSER
     printf("Starting block. Length=%i.\n",remains);
    #endif
   }
  } 
  else
  {
   /* We are inside the block, the c is a data byte. */
   switch (accept_byte(c))
   {
    case 1:
    fprintf(stderr,"Unable to accept byte of GIF data.\n");
    return 1;
    
    case 2:
    free_big_buffer();
    free_line_buffers(); /* No more needed */
    return 2; /* No more raster data wanted. */
   }
   remains--;
  }
  break;
  
  case 7: /* Reading a byte after '!' in GIF */
  tlen=0;
  remains=0;
  state=8;
  break;

  case 8: /* Skipping ignored blocks in GIF */
  if (!remains)
  {
   /* Byte count awaited */
   if (!c)
   {
    /* End. :-) */
    state=2; /* Go and wait for ',' */
    tlen=0;
    break;
   }
   remains=c;
   #ifdef DG_IMAGE_PARSER
    printf("Skipping %i bytes.\n",remains);
   #endif
  }
  else
  {
   remains--;
  }
  break;
  
  case 9: /* Reading header in PCX -- start of PCX code */
  tbuf[tlen]=c;
  tlen++;
  if (tlen<128) break;
  #ifdef DG_IMAGE_PARSER
   printf("PCX header read.\n");
  #endif
  {
   int xmax,xmin,ymax,ymin;
   xmin=tbuf[4]+(tbuf[5]<<8);
   ymin=tbuf[6]+(tbuf[7]<<8);
   xmax=tbuf[8]+(tbuf[9]<<8);
   ymax=tbuf[10]+(tbuf[11]<<8);
   im_width=xmax-xmin+1;
   im_height=ymax-ymin+1;
   if (im_width<=0||im_height<=0)
   {
    if (issue_image_msg("Invalid PCX image dimensions."))
    {
     fprintf(stderr,"Unable to issue message about invalid image"
      " dimensions.\n");
     return 1;
    }
    return 2;
   }
   take_image_line();
  }
  im_bpp=tbuf[3];
  im_planes=tbuf[65];
  im_bpl=tbuf[66]+(tbuf[67]<<8);
  if (!im_bpp||!im_planes||!im_bpl)
  {
   if (issue_image_msg("Zero im_bpp, im_bpl or im_planes in PCX: invalid format."))
   {
    fprintf(stderr,"Unable to issue message about invalid format.\n");
    return 1;
   }
   return 2; 
  }
  if (setup_16_pcx_palette(tbuf+16))
  {
   fprintf(stderr,"Unable to setup 16-color PCX palette.\n");
   return 1;
  }
  #ifdef DG_IMAGE_PARSER
   printf("Starting palette set up.\n");
  #endif
  alloc_line_buffers(im_width);
  alloc_raw_line(im_bpl*im_planes);
  #ifdef DG_IMAGE_PARSER
   printf("raw_line allocated.\n");
  #endif
  if (tbuf[2]!=1)
  {
   free_raw_line();
   free_line_buffers();
   if (issue_image_msg("Unknown PCX encoding."))
   {
    fprintf(stderr,"Unable to issue error message about unknown PCX"
     " encoding.\n");
    return 1;
   }
   return 2; /* Stop sending data. */
  }
  palette_present=(tbuf[1]==5&&im_bpp==8&&im_planes==1);
  if (palette_present)
  { 
   #ifdef DG_IMAGE_PARSER
    fprintf(stderr,"alloc_big_buffer(%d)\n",im_height*im_bpl*im_planes);
   #endif
   alloc_big_buffer(im_height*im_bpl*im_planes);
  }
  tlen=0;
  xoff=0;
  yoff=0;
  state=10; /* Reading PCX pixels */
  break;

  case 12: /* Reading the PCX palette at the end of the file */
  if (c==256)
  {
   accept_raw_pcx_palette(); /* Takes cyclic and cyclic_pos and sets up the 
                                 palette.
			      */
   free(cyclic);
   cyclic=NULL;
   if (redraw_pcx())
   {
    free_big_buffer();
    free_raw_line();
    free_line_buffers();
    fprintf(stderr,"Can't finish PCX.\n");
    return 1;
   }
   free_big_buffer();
   free_raw_line();
   free_line_buffers();
   return 2; /* Stop sending data -- there is nothing interesting behind the
                palette */
  }
  else
  {
   cyclic[cyclic_pos]=c;
   cyclic_pos++;
   if (cyclic_pos>=769)
   {
    cyclic_enough=1;
    cyclic_pos=0;
   }
  }
  break;
  
 }
 return 0;
}

/* Gets a raw row in the "raw_line" and dithers it using the 3-3-2 bit
color palette. The top three bits are red, the middle three bits are green,
and the bottom two bits are blue. */
int
flush_jpeg_line(void)
{
 int r,g,b,a;

 yoff=cinfo.output_scanline;
 for (a=0;a<line_width;a++)
 {
  if (cinfo.output_components>=3)
  {
   r=raw_line[a*cinfo.output_components];
   g=raw_line[a*cinfo.output_components+1];
   b=raw_line[a*cinfo.output_components+2];
  }
  else
  {
   r=raw_line[a*cinfo.output_components];
   g=raw_line[a*cinfo.output_components];
   b=raw_line[a*cinfo.output_components];
  }
  actual_line[a*3]=BTOP(r);
  actual_line[a*3+1]=BTOP(g);
  actual_line[a*3+2]=BTOP(b);
 }
 if (apply_gamma(),dither_and_flush()) goto err;
 yoff++;
 return 0;
 err:
 fprintf(stderr,"Can't flush JPEG line.\n");
 return 1;
}

METHODDEF(void) /* Only for the sake of libjpeg */
nop(j_decompress_ptr cinfo)
{
 #ifdef DG_IMAGE_PARSER
  printf("nop\n");
 #endif
}

METHODDEF(boolean)
my_fill_input_buffer(j_decompress_ptr cinfo)
{
 #ifdef DG_IMAGE_PARSER
  printf("my_fill_input_buffer\n");
 #endif
 return FALSE; /* We utilize I/O suspension (or emulsion? ;-) ) */
}

METHODDEF(void)
my_skip_input_data(j_decompress_ptr cinfo,long num_bytes)
{
 #ifdef DG_IMAGE_PARSER
  printf("my_skip_input_data(%li).\n",num_bytes);
 #endif
 if (num_bytes>cinfo->src->bytes_in_buffer)
 {
  /* We have to enter skipping state */
  cinfo->src->next_input_byte+=cinfo->src->bytes_in_buffer;
  skip_bytes=num_bytes-cinfo->src->bytes_in_buffer;
  cinfo->src->bytes_in_buffer=0;
 }
 else
 {
  /* We only pull out some bytes from buffer. */
  cinfo->src->next_input_byte+=num_bytes;
  cinfo->src->bytes_in_buffer-=num_bytes;
 }
}

METHODDEF(void)
my_error_exit(j_common_ptr cinfo)
{
 #ifdef DG_ERR
  printf("Entering my_error_exit.\n");
 #endif
 (*cinfo->err->output_message) (cinfo);
 #ifdef DG_ERR
  printf("Leaving my_error_exit by longjmp(2).\n");
 #endif
 longjmp(jerr.setjmp_buffer,2);
}

METHODDEF(void)
my_output_message(j_common_ptr cinfo)
{
 char ch;
 struct jpeg_error_mgr *err=cinfo->err;
 enum{BLEN=256};
 unsigned char *buf;
 int msg_code=err->msg_code;
 const char * msgptr;
 const char *msgtext=NULL;
 int isstring=0;
 
 #ifdef DG_ERR
  printf("Entering my_output_message.\n");
 #endif
 buf=malloc(BLEN);
 if (!buf)
 {
  fprintf(stderr,"Unable to malloc a buffer for outputting a JPEG error "
   "message.\n");
  #ifdef DG_ERR
   printf("Leaving my_output_message by longjmp(1).\n");
  #endif
  longjmp(jerr.setjmp_buffer,1);
 }
 if (msg_code>0&&msg_code<err->last_jpeg_message)
 {
  msgtext=err->jpeg_message_table[msg_code];
 }
 if (!msgtext)
 {
  err->msg_parm.i[0] = msg_code;
  msgtext = err->jpeg_message_table[0];
 }
 
 /* Check for string parameter, as indicated by %s in the message text */
 msgptr = msgtext;
 while ((ch = *msgptr++) != '\0')
 {
  if (ch == '%') 
  {
   if (*msgptr == 's') isstring = 1;
   break;
  }
 } 

 /* Format the message into the passed buffer */
 if (isstring)
  snprintf(buf,BLEN,msgtext, err->msg_parm.s);
 else
  snprintf(buf,BLEN,msgtext,
            err->msg_parm.i[0], err->msg_parm.i[1],
            err->msg_parm.i[2], err->msg_parm.i[3],
            err->msg_parm.i[4], err->msg_parm.i[5],
            err->msg_parm.i[6], err->msg_parm.i[7]);
 if (issue_image_msg(buf))
 {
  fprintf(stderr,"Unable to issue_image_msg JPEG error message.\n");
  free(buf);
  #ifdef DG_ERR
   printf("Leaving my_output_message by longjmp(2).\n");
  #endif
  longjmp(jerr.setjmp_buffer,1);
 }
 free(buf);
}

void
png_error_fn(png_structp png_ptr, png_const_charp error_msg)
{
 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"png_error_fn(%s)\n",error_msg);
 #endif
 if (issue_image_msg(error_msg))
 {
  #ifdef DG_IMAGE_PARSER
   fprintf(stderr,"issue_image_msg failed.\n");
  #endif
  error_occured=1;
 }
 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"!png_error_fn -- longjump.\n");
 #endif
 longjmp(png_ptr->jmpbuf,1);
}

void
png_warning_fn(png_structp png_ptr, png_const_charp warning_msg)
{
 if (issue_image_msg(warning_msg))
 {
  error_occured=1;
  longjmp(png_ptr->jmpbuf,1); 
 }
}

/* Allocs line buffers */
void
info_callback(png_structp png_ptr, png_infop info)
{
 int color_type;
 double gamma;
 png_color_16 my_background={0,0,0,0,0};
 png_color_16p image_background;
 int bit_depth;
 int intent;

 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"info_callback\n");
 #endif
 bit_depth=png_get_bit_depth(png_ptr,info_ptr);
 color_type=png_get_color_type(png_ptr,info_ptr);
 im_width=png_get_image_width(png_ptr,info_ptr);
 im_height=png_get_image_height(png_ptr,info_ptr);
 png_set_interlace_handling(png_ptr);
 png_set_packing(png_ptr);
 if (color_type == PNG_COLOR_TYPE_PALETTE)
  png_set_expand(png_ptr);
 if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
  png_set_expand(png_ptr);
 if (png_get_sRGB(png_ptr, info_ptr, &intent))
  png_set_sRGB(png_ptr, info_ptr, intent);
 else
 {              
  if (png_get_gAMA(png_ptr, info_ptr, &gamma))
  {
   png_set_gamma(png_ptr,1,image_gamma<=0?gamma:image_gamma);
   #ifdef DG_GAMMA
    fprintf(stderr,"gAMA=%s%f.\n",image_gamma<=0?"":" (forced) ",image_gamma<=0?gamma:image_gamma);
   #endif
  }
  else
  { 
   png_set_gamma(png_ptr,1, image_gamma<=0?1:image_gamma);
   #ifdef DG_GAMMA
    fprintf(stderr,"User-set gamma %f\n",image_gamma<=0?.5:image_gamma);
   #endif
  }
 }

  if (png_get_bKGD(png_ptr, info_ptr, &image_background))
  {   
    #ifdef DG_IMAGE_PARSER
     fprintf(stderr,"Using image background\n"); 
    #endif 
     png_set_background(png_ptr, image_background,
                        PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
  }
  else
     png_set_background(png_ptr, &my_background,
                        PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);
 
 png_read_update_info(png_ptr,info_ptr);                 
            
 if (PNG_INTERLACE_NONE!=png_get_interlace_type(png_ptr,info_ptr))
 {
  alloc_big_buffer(png_get_rowbytes(png_ptr,info_ptr)*im_height);
  interl_dist=1;
 }
 else
 {
  interl_dist=0;
 }
 alloc_line_buffers(im_width);
 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"%ld bytes per pixel\n",png_get_rowbytes(png_ptr,info_ptr)
   /im_height);
 #endif
 take_image_line();
 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"!info_callback\n");
 #endif
}

void
png_row_to_actual(unsigned char *d)
{
 int x;
 int depth;
 int channels,bpl;
 
 depth=png_get_bit_depth(png_ptr,info_ptr);
 channels=png_get_channels(png_ptr,info_ptr);
 #ifdef GD_IMAGE_PARSER
  fprintf(stderr,"channels %d\n",channels);
 #endif
 bpl=png_get_rowbytes(png_ptr,info_ptr);
 #ifdef DG_IMAGE_PARSER
  fprintf(stderr,"Bytes per pixel %d\n",bpl/im_width);
 #endif
 if (channels==1)
 {
  #ifdef DG_IMAGE_PARSER
   fprintf(stderr,"Gray\n");
  #endif
  if (depth==16)
   for (x=0;x<im_width;x++,d+=2)
   {
    int val=WTOP((*d<<8)+d[1]);
    actual_line[3*x]=val;
    actual_line[3*x+1]=val;
    actual_line[3*x+2]=val;
   }
  else
   for (x=0;x<im_width;x++,d++)
   {
    int val=BTOP(*d);
    actual_line[3*x]=val;
    actual_line[3*x+1]=val;
    actual_line[3*x+2]=val;
   }
 }
 else if (channels==3)
 {
  #ifdef DG_IMAGE_PARSER
   fprintf(stderr,"RGB\n");
  #endif
  if (depth==16)
   for (x=0;x<im_width*3;x++,d+=2)
    actual_line[x]=WTOP((*d<<8)+d[1]);
  else
   for (x=0;x<im_width*3;x++,d++)
    actual_line[x]=BTOP(*d);
 }
 else
 {
  fprintf(stderr,"PNG library setup error: channels=%d\n",channels);
 }
}

void
row_callback(png_structp pnt_ptr, png_bytep new_row, png_uint_32 row_num,
 int pass)
{
 int rb=png_get_rowbytes(png_ptr,info_ptr);
 
 if (!row_num)
 {
  zero_error_line(); /* Restart dithering on top of the image */
 }
 yoff=row_num;
 if (interl_dist)
 {
  png_progressive_combine_row(png_ptr,big_buffer+rb*yoff,new_row);
  png_row_to_actual(big_buffer+rb*yoff);
 }
 else
  png_row_to_actual(new_row);
 dither_and_flush();
}

void
end_callback( png_structp png_ptr, png_infop info)
{
 png_complete=1;
}

/* Return value:
   0=ok
   1=error
   2=stop sending image data
*/
int
image_parser(int command, void *params)
{
 volatile struct block_desc b;
 unsigned char *ptr,*end;
 int c;
 
 switch(command)
 {
  case IP_NEW_SCREEN:
  {
   int type;

   type=*(int *)params;
   #ifdef DG_IMAGE_PARSER
    printf("IP_NEW_SCREEN,%i.\n",type);
   #endif
   im_height=0;
   usable_line=0;
   image_line=0;
   switch(type)
   {
    case 0: /* Initialize for GIF */
    skip_bytes=0;
    tlen=0;
    state=0; /* Reading GIF signature */
    break;

    case 1: /* Initialize for PCX */
    skip_bytes=0;
    tlen=0;
    state=9; /* Reading PCX signature */
    break;

    case 2: /* Initialize for JPEG */
    skip_bytes=0;
    tlen=0;
    state=13;
    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit=my_error_exit;
    jerr.pub.output_message=my_output_message;
    #ifdef DG_IMAGE_PARSER
     printf("I got behind jpeg_std_error.\n");
    #endif
    jpeg_create_decompress(&cinfo);
    #ifdef DG_IMAGE_PARSER
     printf("I got behind jpeg_create_decompress.\n");
    #endif
    jpeg_stdio_src(&cinfo,stdin);
    #ifdef DG_IMAGE_PARSER
     printf("I got behind jpeg_stdio_src.\n");
    #endif
    cinfo.src->init_source=nop;
    cinfo.src->fill_input_buffer=my_fill_input_buffer;
    cinfo.src->skip_input_data=my_skip_input_data;
    cinfo.src->resync_to_restart=jpeg_resync_to_restart;
    cinfo.src->term_source=nop;
    cinfo.src->bytes_in_buffer=0;
    cinfo.src->next_input_byte=NULL;
    #ifdef DG_IMAGE_PARSER
     printf("JPEG initialization complete.\n");
    #endif
    break;

    case 3: /* Initialize for PNG */
    tlen=0;
    skip_bytes=0;
    state=19;
    png_complete=0;
    
    png_ptr=png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,png_error_fn,
     png_warning_fn);
    if (!png_ptr)
    {
     fprintf(stderr,"Can't create png_ptr.\n");
     return 1;
    }
    
    info_ptr=png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
     png_destroy_read_struct(&png_ptr,(png_infopp)NULL,(png_infopp)NULL);
     fprintf(stderr,"Can't create info_ptr for PNG\n");
     return 1;
    }
    
    #define SETJMP error_occured=0;\
    if (setjmp(png_ptr->jmpbuf))\
    {\
     if (error_occured)\
     {\
      fprintf(stderr,"Fatal error occured while reading PNG.\n");\
      return 1;\
     }\
     \
     png_destroy_read_struct(&png_ptr,&info_ptr,NULL);\
     free_line_buffers();\
     free_big_buffer();\
     return 2; /* Stop sending data */\
    }

    SETJMP
    png_set_progressive_read_fn(png_ptr,NULL,info_callback,row_callback,
     end_callback);
    break;
   
   }
  }
  break;
  
  case IP_PARSE_BLOCK:
  b=*(struct block_desc*)params;
  #ifdef DG_IMAGE_PARSER
   printf("I got into IP_PARSE_BLOCK with block length=%i.\n",b.number);
  #endif
  if (!b.number)
  {
   break;
  }
  if (skip_bytes)
  {
   #ifdef DG_IMAGE_PARSER
    printf("JPEG: skipping bytes.\n");
   #endif
   if (skip_bytes>=b.number)
   {
    /* We have to throw out all the buffer */
    skip_bytes-=b.number;
    break; /* Return from the image_parser() */
   }
   else
   {
    /* We only shorten the buffer */
    b.ptr+=skip_bytes;
    b.number-=skip_bytes;
    skip_bytes=0;
   }
  }
  if (state>=13&&state<=18)
  {
   switch (setjmp(jerr.setjmp_buffer))
   {
    case 2: /* Normal error */
    #ifdef DG_ERR
     printf("setjmp(2).\n");
    #endif
    jpeg_destroy_decompress(&cinfo);
    #ifdef DG_IMAGE_PARSER
     printf("Destroy done.\n");
    #endif
    free_line_buffers();
    free_raw_line();
    #ifdef DG_IMAGE_PARSER
     printf("Raw_line done.\n");
    #endif
    free(jdata);
    jdata=NULL;
    #ifdef DG_IMAGE_PARSER
     printf("jdata done.\n");
    #endif
    return 2; /* Stop sending data */
    
    case 1: /* My fatal error */
    #ifdef DG_IMAGE_PARSER
     printf("setjmp(1).\n");
    #endif
    return 1;
    break;
   }

   /* We have to add the arriving data to the data still left in the buffer */
   #ifdef DG_IMAGE_PARSER
    printf("Adding data for JPEG.\n");
   #endif
   if (jdata)
   {
    memcpy(jdata,cinfo.src->next_input_byte,cinfo.src->bytes_in_buffer);
   }
   jdata=realloc(jdata,cinfo.src->bytes_in_buffer+b.number);
   if (!jdata)
   {
    fprintf(stderr,"image_parser (JPEG): realloc ->%i failed.\n",cinfo.src
     ->bytes_in_buffer+b.number);
    return 1;
   }
   memcpy(jdata+cinfo.src->bytes_in_buffer,b.ptr,b.number);
   cinfo.src->next_input_byte=jdata;
   cinfo.src->bytes_in_buffer+=b.number;
   #ifdef DG_IMAGE_PARSER
    printf("After the preparation of IP_PARSE_BLOCK, cinfo.src->bytes_in_buffer"
     "=%i.\n",cinfo.src->bytes_in_buffer);
    #endif
  }
  switch(state)
  {
   case 13:
   /* Reading JPEG header */
   #ifdef DG_IMAGE_PARSER
    printf("case 13:\n");
   #endif
   if (JPEG_SUSPENDED==jpeg_read_header(&cinfo,TRUE))
   {
    #ifdef DG_IMAGE_PARSER
     printf("jpeg_read_header=JPEG_SUSPENDED.\n");
    #endif
    break; /* We will repeat it next time */
   }
   cinfo.buffered_image=TRUE;
   state=14; /* jpeg_start_decompress */
   /* break; left out intentionally */
   
   case 14:
   #ifdef DG_IMAGE_PARSER
    printf("case 14:\n");
   #endif
   if (jpeg_start_decompress(&cinfo)==FALSE)
   {
    #ifdef DG_IMAGE_PARSER
     printf("jpeg_start_decompress=FALSE.\n");
    #endif
    break;
   }
   im_width=cinfo.output_width;
   im_height=cinfo.output_height;
   take_image_line();
   alloc_raw_line(im_width*cinfo.output_components);
   alloc_line_buffers(im_width);
   #ifdef DG_IMAGE_PARSER
    printf("Allocated raw_line to %i bytes. cinfo.output_components=%i, "
     "cinfo.output_width=%i.\n",cinfo.output_components*cinfo.output_width
     ,cinfo.output_components,cinfo.output_width);
   #endif
   new_scan:
   state=17;
   /* break intentionally left out */

   case 17:
   if (FALSE==jpeg_start_output(&cinfo,cinfo.input_scan_number))
   {
    /* Suspended. */
    break;
   }
   zero_error_line();
   state=15; /* jpeg-read_scanlines */
   /* break; left out intentionally */
   
   case 15:
   #ifdef DG_IMAGE_PARSER
    printf("case 15:\n");
   #endif
   while (cinfo.output_scanline<cinfo.output_height)
   {
    unsigned char *row=raw_line;
    /* Scanlines remain to be read */
    if (jpeg_read_scanlines(&cinfo,&row,1))
    {
     /* We read a scanline */
     if (flush_jpeg_line())
     {
      fprintf(stderr,"image_parser: unable to flush_jpeg_line().\n");
      return 1;
     }
    } 
    else
    {
     /* We are suspended and we want more data */
     goto susp0;
    }
   }
   state=18;
   /* break intentionally left out. */
   
   case 18:
   if (FALSE==jpeg_finish_output(&cinfo))
   {
    /* Suspended */
    break;
   }
   if (!jpeg_input_complete(&cinfo))
   {
    /* Some more scans awaited... */
    goto new_scan;
   }
   /* All scanlines read */
   free_raw_line();
   free_line_buffers();
   state=16; /* jpeg_finish_decompress */
   /* break; intentionally left out */

   case 16:
   if (FALSE==jpeg_finish_decompress(&cinfo))
   {
    break;
   }
   jpeg_destroy_decompress(&cinfo);
   free(jdata);
   jdata=NULL;
   return 2; /* We have done all the JPEG image! :-) */
   susp0:
   break;
   
   case 19: /* This is the beginning of PNG decoding -- processing data */
   #ifdef DG_IMAGE_PARSER
    fprintf(stderr,"case 19: calling png_process_data\n");
   #endif
   SETJMP
   png_process_data(png_ptr,info_ptr,b.ptr,b.number);
   #ifdef DG_IMAGE_PARSER
    fprintf(stderr,"png_process_data returned\n");
   #endif
   if (png_complete)
   {
    /* The end of the image */
    png_destroy_read_struct(&png_ptr,&info_ptr,NULL);
    free_big_buffer();
    free_line_buffers();
    return 2; /* Stop sending data */
   }
   break;
  
   state_10:
   case 10: /* Reading ordinary byte or run length in PCX */
   /* We have the chunk of data in block.ptr and block.number. */
   {
    ptr=b.ptr;
    end=ptr+b.number;
    
    while(ptr<end)
    {
     c=*ptr++;
     if (c<192)
     {
      /* Normal */
      *raw_line_ptr++=c;
      if (raw_line_ptr>=raw_line_end) goto eol;
     }
     else
     {
      /* Run length */
      run_length=c&63;
      if (ptr>=end)
      {
       state=11;
       return 0;
      }
      pcx_run_length: /* Resume */
      c=*ptr++;
      while(run_length)
      {
       int remains=run_length;
       if (raw_line_ptr+remains>raw_line_end)
        remains=raw_line_end-raw_line_ptr;
       memset(raw_line_ptr,c,remains);
       raw_line_ptr+=remains;
       run_length-=remains;
       
       if (raw_line_ptr>=raw_line_end)
       { 
        eol:
        switch(accept_pcx_eol())
        {
         case 1:
         free_raw_line();
         free_line_buffers();
         free_big_buffer();
         fprintf(stderr,"Unable to accept PCX byte.\n");
         return 1;
     
         case 2:
         state=12;
         b.ptr=ptr;
         b.number=end-ptr;
         goto default_label;
     
         case 3:
         return 2; /* Really everyting is done */
        }
       }
      }
     }
    }
   }
   break;
  
   case 11: 
   state=10;
   ptr=b.ptr;
   end=ptr+b.number;
   goto pcx_run_length;
  
   default:
   default_label:
   /* Something else than JPEG */
   while(b.number)
   {
    switch(accept(*(b.ptr)))
    {
     case 1:
     fprintf(stderr,"image_parser: can't accept incoming char.\n");
     return 1;
 
     case 2:
     return 2;
     /* No next data are needed */
    }
    b.ptr++;
    b.number--;
    if (state==10) goto state_10;
   }
   break;
  }
  break;
  
  case IP_EOF:
  #ifdef DG_IMAGE_PARSER
   printf("I got IP_EOF.\n");
  #endif
  if (state>=13&&state<=19)
  {
   /* JPEG or PNG */
   free_line_buffers();
   free_big_buffer();
   free_raw_line();
   issue_image_msg("Image is incomplete.");
   return 2; /* Stop sending data */
  }
  else
   switch(accept(256))
   {
    case 1:
    fprintf(stderr,"image_parser: can't do accept(EOF).\n");
    return 1;
  
    case 2:
    return 2; /* Stop sending data */
   }
  break;

  case IP_CALIBRATE:
  {
   int x;
   int height=ysize-font_height;
   double yd;
   
   if (height<64) height=64;
   alloc_line_buffers(xsize);
   for (yoff=0;yoff<height;yoff++)
   {
    yd=(double)yoff/(height-1);
    if (yd<.333333333333333333333333)
    {
     for (x=0;x<line_width;x++)
     {
      int i=x*8/line_width;
      actual_line[3*x]=(i==0||i==1||i==6||i==7)?(i&1?(yoff&1?0:PRESCALE):PRESCALE/2):0;
      actual_line[3*x+1]=(i==2||i==3||i==6||i==7)?(i&1?(yoff&1?0:PRESCALE):PRESCALE/2):0;
      actual_line[3*x+2]=(i==4||i==5||i==6||i==7)?(i&1?(yoff&1?0:PRESCALE):PRESCALE/2):0;
     }
    }
    else if (yd<.666666666666666666666666666666)
    {
     for (x=0;x<xsize;x++)
     {
      int i=x*8/line_width;
      double d=(yd-.33333333333333333333333333333333)*3;
      d=pow(d,3);
      actual_line[3*x]=i&2?PRESCALE*d:0;
      actual_line[3*x+1]=i&4?PRESCALE*d:0;
      actual_line[3*x+2]=i&1?PRESCALE*d:0;
     }
    }
    else
    {
     for (x=0;x<line_width;x++)
     {
      int vv;
      double v=(double)x/(line_width-1);
      v=pow(v,3);
      vv=RND(PRESCALE*v);
      actual_line[3*x]=vv;
      actual_line[3*x+1]=vv;
      actual_line[3*x+2]=vv;
     }
    }
   dither_and_flush();
   }
   free_line_buffers();
  }
  break;
 }
 return 0;
}
